/*====================================================================================
Class Name  : ARApp
Authors	    : Grant Clark, Ewan Duncan
Description : Main class of the program.  Manages the program states and renders
			  all sprites and meshes.  Adapted by Ewan Duncan from Grant Clark's
			  ARApp class.
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "ar_app.h"

#include <system/platform.h>
#include <graphics/font.h>
#include <sony_sample_framework.h>
#include <sony_tracking.h>

#include "marker_select_state.h"

void *graphicsAlloc(SceKernelMemBlockType type, uint32_t size, uint32_t alignment, uint32_t attribs, SceUID *uid);
void graphicsFree(SceUID uid);

/*------------------------------------------------------------------------------------
CONSTRUCTORS AND DESTRUCTORS
------------------------------------------------------------------------------------*/
ARApp::ARApp(abfw::Platform& platform) :
	Application(platform),
	input_manager_(NULL),
	sprite_renderer_(NULL),
	font_(NULL),
	renderer_3d_(NULL)
{
}

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Init
Purpose : Initialises the application.
====================================================================================*/
void ARApp::Init()
{
	// create the input manager and the renderers
	input_manager_ = platform_.CreateSonyControllerInputManager();
	sprite_renderer_ = platform_.CreateSpriteRenderer();
	renderer_3d_ = platform_.CreateRenderer3D();

	// initialise the font
	InitFont();

	// initialise sony framework
	sampleInitialize();
	smartInitialize();

	// reset marker tracking
	AppData* dat = sampleUpdateBegin();
	smartTrackingReset();
	sampleUpdateEnd(dat);

	// set up the orthographic frustum
	orthographic_projection_.OrthographicFrustumVita(-1.f, 1.f, -1.f, 1.f, -1.f, 1.f);

	// define the image scale factor for the camera feed
	vertical_image_scale_factor_ = (960.f / 544.f) / (640.f / 480.f);

	// set up the field of view
	fov_projection_matrix_.PerspectiveFovVita(SCE_SMART_IMAGE_FOV, SCE_SMART_IMAGE_WIDTH / SCE_SMART_IMAGE_HEIGHT, 0.001f, 2.f);

	// create a scale factor using the image scale factor
	abfw::Matrix44 scale_factor;
	scale_factor.Scale(abfw::Vector3(1.f, vertical_image_scale_factor_, 1.f));

	// use the scale factor to scale the fielf o
	fov_projection_matrix_ = fov_projection_matrix_ * scale_factor;

	// initialise a sprite for rendering the camera feed
	camera_image_sprite_.set_width(2.f);
	camera_image_sprite_.set_height(2.f * vertical_image_scale_factor_);
	camera_image_sprite_.set_position(0.f, 0.f, 1.f);
	camera_image_sprite_.set_texture(&texture_map_);

	// start the program in the marker select state
	state_manager_.ChangeState(new MarkerSelectState(&platform_));
}

/*====================================================================================
Name    : CleanUp
Purpose : Performs any necessary memory cleanup
====================================================================================*/
void ARApp::CleanUp()
{
	smartRelease();
	sampleRelease();

	CleanUpFont();
	delete sprite_renderer_;
	sprite_renderer_ = NULL;

	delete renderer_3d_;
	renderer_3d_ = NULL;

	delete input_manager_;
	input_manager_ = NULL;

	state_manager_.Clear();
}

/*====================================================================================
Name    : Update
Purpose : The main update loop for the program.  
====================================================================================*/
bool ARApp::Update(float frame_time)
{
	fps_ = 1.0f / frame_time;

	AppData* dat = sampleUpdateBegin();

	// use the tracking library to try and find markers
	smartUpdate(dat->currentImage);
	
	// update the input manager and get the controller from it
	input_manager_->Update();
	const abfw::SonyController* controller = input_manager_->GetController(0);

	// update the top program state in the vector of states
	state_manager_.states_.back()->Update(frame_time, controller, &state_manager_);

	sampleUpdateEnd(dat);

	return true;
}

/*====================================================================================
Name    : Render
Purpose : Renders all the objects for the program.  
====================================================================================*/
void ARApp::Render()
{
	AppData* dat = sampleRenderBegin();

	// set the projection matrix for the sprite renderer
	sprite_renderer_->set_projection_matrix(orthographic_projection_);

	// start drawing with the sprite renderer
	sprite_renderer_->Begin(true);

	// render the camera feed
	if (dat->currentImage)
	{
		texture_map_.set_texture(dat->currentImage->tex_yuv);
		sprite_renderer_->DrawSprite(camera_image_sprite_);
	}

	// set the projection and view matrix for the 3D renderer
	renderer_3d_->set_projection_matrix(fov_projection_matrix_);
	abfw::Matrix44 view_matrix;
	view_matrix.SetIdentity();
	renderer_3d_->set_view_matrix(view_matrix);

	// begin rendering 3D meshes, don't clear the frame buffer
	renderer_3d_->Begin(false);

	// render all objects of the top program state in the vector of states
	state_manager_.states_.back()->Render(sprite_renderer_, renderer_3d_);
	
	// stop rendering with the sprite renderer and 3D renderer
	sprite_renderer_->End();
	renderer_3d_->End();

	RenderOverlay();

	sampleRenderEnd();
}

/*====================================================================================
Name    : RenderOverLay
Purpose : Renders a 2D hud over other objects
====================================================================================*/
void ARApp::RenderOverlay()
{
	abfw::Matrix44 proj_matrix2d;
	proj_matrix2d.OrthographicFrustumVita(0.0f, platform_.width(), 0.0f, platform_.height(), -1.0f, 1.0f);
	sprite_renderer_->set_projection_matrix(proj_matrix2d);
	sprite_renderer_->Begin(false);
	DrawHUD();
	sprite_renderer_->End();
}

/*====================================================================================
Name    : InitFont
Purpose : Intialises the font.
====================================================================================*/
void ARApp::InitFont()
{
	font_ = new abfw::Font();
	font_->Load("comic_sans", platform_);
}

/*====================================================================================
Name    : CleanUpFont
Purpose : Deletes the font.  
====================================================================================*/
void ARApp::CleanUpFont()
{
	delete font_;
	font_ = NULL;
}

/*====================================================================================
Name    : DrawHUD
Purpose : Renders any hud objects such as the fps.
====================================================================================*/
void ARApp::DrawHUD()
{
	if(font_)
	{
		font_->RenderText(sprite_renderer_, abfw::Vector3(850.0f, 510.0f, -0.9f), 1.0f, 0xffffffff, abfw::TJ_LEFT, "FPS: %.1f", fps_);
	}
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
