/*====================================================================================
Class Name  : GameObject
Authors	    : Ewan Duncan
Description : Class that represents an object in the game that gets updated and
			  rendered.
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "game_object.h"

#include <system/platform.h>
#include <graphics/renderer_3d.h>
#include <graphics/image_data.h>
#include <assets/png_loader.h>
#include <assets/obj_loader.h>
#include <graphics/mesh.h>
#include <graphics/primitive.h>
#include <maths/sphere.h>
#include <sony_sample_framework.h>
#include <maths/math_utils.h>

/*------------------------------------------------------------------------------------
CONSTRUCTORS AND DESTRUCTORS
------------------------------------------------------------------------------------*/
GameObject::GameObject()
{
	transform_.SetIdentity();
	local_transform_.SetIdentity();
}

GameObject::~GameObject()
{
	model_.Release();
}

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Render
Purpose : Renders the object's mesh.  
====================================================================================*/
void GameObject::Render(abfw::Renderer3D* renderer_3d)
{
	renderer_3d->DrawMesh(*this);
}

/*====================================================================================
Name    : SetMeshToModel
Purpose : Loads in a 3D model and sets the object mesh to match the model's mesh.
====================================================================================*/
abfw::Mesh* GameObject::CreateModelMesh(const char* filename, const abfw::Platform& platform, const abfw::Aabb aabb)
{
	// load in 3D model
	abfw::OBJLoader obj_loader;
	obj_loader.Load(filename, platform, model_);
	
	// set the aabb and the mesh on the object
	model_.mesh()->set_aabb(aabb);
	
	// return the created mesh
	return model_.mesh();
}

/*====================================================================================
Name    : SetMeshToCube
Purpose : Creates a cube mesh for the mesh of the gameobject.
====================================================================================*/
abfw::Mesh* GameObject::CreateCubeMesh(const char* texture_filename, const abfw::Platform& platform)
{
	// create a mesh
	abfw::Mesh* mesh = platform.CreateMesh();

	// define half the mesh size for setting up the vertices
	const float half_mesh_size = 0.5f;

	// set up the vertices using half the mesh size
	const abfw::Mesh::Vertex vertices[] = {
		{ abfw::Vector3(half_mesh_size, -half_mesh_size, -half_mesh_size), abfw::Vector3(0.577f, -0.577f, -0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(half_mesh_size, half_mesh_size, -half_mesh_size), abfw::Vector3(0.577f, 0.577f, -0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(-half_mesh_size, half_mesh_size, -half_mesh_size), abfw::Vector3(-0.577f, 0.577f, -0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(-half_mesh_size, -half_mesh_size, -half_mesh_size), abfw::Vector3(-0.577f, -0.577f, -0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(half_mesh_size, -half_mesh_size, half_mesh_size), abfw::Vector3(0.577f, -0.577f, 0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(half_mesh_size, half_mesh_size, half_mesh_size), abfw::Vector3(0.577f, 0.577f, 0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(-half_mesh_size, half_mesh_size, half_mesh_size), abfw::Vector3(-0.577f, 0.577f, 0.577f), abfw::Vector2(0.0f, 0.0f) },
		{ abfw::Vector3(-half_mesh_size, -half_mesh_size, half_mesh_size), abfw::Vector3(-0.577f, -0.577f, 0.577f), abfw::Vector2(0.0f, 0.0f) } };
	
	// initialise the vertex buffer
	mesh->InitVertexBuffer(platform, static_cast<const void*>(vertices), sizeof(vertices) / sizeof(abfw::Mesh::Vertex), sizeof(abfw::Mesh::Vertex));

	// define half the size of an axis aligned bounding box
	float half_aabb_size = half_mesh_size * 0.95f;
	
	// create an aabb and use the half aabb size to set it's min and max vertex
	abfw::Aabb aabb;
	aabb.set_min_vtx(abfw::Vector3(-half_aabb_size, -half_aabb_size, -half_aabb_size));
	aabb.set_max_vtx(abfw::Vector3(half_aabb_size, half_aabb_size, half_aabb_size));

	// set the aabb of the mesh
	mesh->set_aabb(aabb);

	// allocate primitives and get the first primitive
	mesh->AllocatePrimitives(1);
	abfw::Primitive* primitive = mesh->GetPrimitive(0);

	// set up the indices
	const UInt32 indices[] = {
		// Back
		0, 1, 2,
		2, 3, 0,
		// Front
		6, 5, 4,
		4, 7, 6,
		// Left
		2, 7, 3,
		2, 6, 7,
		// Right
		0, 4, 1,
		5, 1, 4,
		// Top
		6, 2, 1,
		5, 6, 1,
		// Bottom
		0, 3, 7,
		0, 7, 4
	};

	// initialise the index buffer and set the primitive type
	primitive->InitIndexBuffer(platform, static_cast<const void*>(indices), sizeof(indices) / sizeof(UInt32), sizeof(UInt32));
	primitive->SetType(abfw::TRIANGLE_LIST);

	// set the texture
	primitive->set_texture(LoadTextureFromPNG(texture_filename, platform));

	// return the created mesh
	return mesh;
}

/*====================================================================================
Name    : LoadTextureFromPNG
Purpose : Loads in a png file and uses it to create a texture.  
====================================================================================*/
abfw::Texture* GameObject::LoadTextureFromPNG(const char* filename, const abfw::Platform& platform)
{
	// create a png loader
	abfw::PNGLoader png_loader;

	// create a structure to hold the image data
	abfw::ImageData image_data;

	// load the png
	png_loader.Load(filename, platform, image_data);

	// check that image data is valid
	if (image_data.image() == NULL)
	{
		return NULL;
	}

	// create the texture from the image data
	abfw::Texture* texture = platform.CreateTexture(image_data);

	// return the texture
	return texture;
}

/*====================================================================================
Name    : MultiplyByMarkerTransform
Purpose : Multiplies the local transform of the game object by the transform of an 
		  ar marker to get the world transform of the object.
====================================================================================*/
void GameObject::MultiplyByMarkerTransform(int marker_index)
{
	// get the marker transform
	abfw::Matrix44 marker_transform;
	sampleGetTransform(
		marker_index,
		&marker_transform);
	
	transform_ = local_transform_ * marker_transform;
}

/*====================================================================================
Name    : GetLocalMarkerTransform
Purpose : Get the local transform of a marker, relative to a reference marker.
====================================================================================*/
abfw::Matrix44 GameObject::GetLocalMarkerTransform(int reference_marker_index, int marker_index)
{
	// get the marker transforms
	abfw::Matrix44 marker1_transform;
	sampleGetTransform(
		reference_marker_index,
		&marker1_transform);
	abfw::Matrix44 marker2_transform;
	sampleGetTransform(
		marker_index,
		&marker2_transform);

	// get the inverse transform of the reference marker
	abfw::Matrix44 marker1_inverse_transform;
	marker1_inverse_transform.AffineInverse(marker1_transform);

	// calculate the local marker transform by multiplying by the inverse transform
	abfw::Matrix44 local_marker_transform = marker2_transform * marker1_inverse_transform;

	// return the result
	return local_marker_transform;
}

/*====================================================================================
Name    : SetPosition
Purpose : Moves the game object to a specified position.  
====================================================================================*/
void GameObject::SetPosition(const float x_position, const float y_position, const float z_position)
{
	local_transform_.SetTranslation(abfw::Vector3(x_position, y_position, z_position));
}

/*====================================================================================
Name    : SetPosition
Purpose : Moves the game object to a specified position.
====================================================================================*/
void GameObject::SetPosition(const abfw::Vector3 position)
{
	local_transform_.SetTranslation(position);
}

/*====================================================================================
Name    : SetXPosition
Purpose : Changes the game object's x coordinate.  
====================================================================================*/
void GameObject::SetXPosition(const float x_position)
{
	local_transform_.SetTranslation(abfw::Vector3(x_position, local_transform_.GetTranslation().y, local_transform_.GetTranslation().z));
}

/*====================================================================================
Name    : SetYPosition
Purpose : Changes the game object's y coordinate.
====================================================================================*/
void GameObject::SetYPosition(const float y_position)
{
	local_transform_.SetTranslation(abfw::Vector3(local_transform_.GetTranslation().x, y_position, local_transform_.GetTranslation().z));
}

/*====================================================================================
Name    : SetZPosition
Purpose : Changes the game object's z coordinate.
====================================================================================*/
void GameObject::SetZPosition(const float z_position)
{
	local_transform_.SetTranslation(abfw::Vector3(local_transform_.GetTranslation().x, local_transform_.GetTranslation().y, z_position));
}

/*====================================================================================
Name    : GetPosition
Purpose : Returns the position of the game object.
====================================================================================*/
abfw::Vector3 GameObject::GetPosition()
{
	return local_transform_.GetTranslation();
}

/*====================================================================================
Name    : MoveDeltaX
Purpose : Moves the game object along the x axis by a specified distance.  
====================================================================================*/
void GameObject::MoveDeltaX(const float x_distance)
{
	local_transform_.SetTranslation(abfw::Vector3(local_transform_.GetTranslation().x + x_distance, local_transform_.GetTranslation().y, local_transform_.GetTranslation().z));
}

/*====================================================================================
Name    : MoveDeltaY
Purpose : Moves the game object along the y axis by a specified distance.
====================================================================================*/
void GameObject::MoveDeltaY(const float y_distance)
{
	local_transform_.SetTranslation(abfw::Vector3(local_transform_.GetTranslation().x, local_transform_.GetTranslation().y + y_distance, local_transform_.GetTranslation().z));
}

/*====================================================================================
Name    : MoveDeltaZ
Purpose : Moves the game object along the z axis by a specified distance.
====================================================================================*/
void GameObject::MoveDeltaZ(const float z_distance)
{
	local_transform_.SetTranslation(abfw::Vector3(local_transform_.GetTranslation().x, local_transform_.GetTranslation().y, local_transform_.GetTranslation().z + z_distance));
}

/*====================================================================================
Name    : SetXRotation
Purpose : Changes the game object's x rotation.
====================================================================================*/
void GameObject::SetXRotation(const float x_rotation)
{
	local_transform_.RotationX(abfw::DegToRad(x_rotation));
}

/*====================================================================================
Name    : SetYRotation
Purpose : Changes the game object's y rotation.
====================================================================================*/
void GameObject::SetYRotation(const float y_rotation)
{
	local_transform_.RotationY(abfw::DegToRad(y_rotation));
}

/*====================================================================================
Name    : SetZRotation
Purpose : Changes the game object's z rotation.
====================================================================================*/
void GameObject::SetZRotation(const float z_rotation)
{
	local_transform_.RotationZ(abfw::DegToRad(z_rotation));
}

/*====================================================================================
Name    : Scale
Purpose : Scales the game object by a specified amount.  
====================================================================================*/
void GameObject::Scale(const float x_scale, const float y_scale, const float z_scale)
{
	abfw::Matrix44 scale_factor;
	scale_factor.Scale(abfw::Vector3(x_scale, y_scale, z_scale));
	local_transform_ = local_transform_ * scale_factor;
}

/*====================================================================================
Name    : AABBCollision
Purpose : Use axis aligned bounding box collision to see if this object is colliding 
		  with another specified game object.
====================================================================================*/
bool GameObject::AABBCollision(GameObject& other_object)
{
	// get the aabbs from the meshes of the two game objects
	abfw::Aabb box1 = mesh()->aabb().Transform(local_transform_);
	abfw::Aabb box2 = other_object.mesh()->aabb().Transform(other_object.get_local_transform());

	// see if the boxes overlap or not and return the result
	return(box1.max_vtx().x > box2.min_vtx().x &&
		box1.min_vtx().x < box2.max_vtx().x &&
		box1.max_vtx().y > box2.min_vtx().y &&
		box1.min_vtx().y < box2.max_vtx().y &&
		box1.max_vtx().z > box2.min_vtx().z &&
		box1.min_vtx().z < box2.max_vtx().z);
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
