/*====================================================================================
Class Name  : LevelEditState
Authors	    : Ewan Duncan
Description : Allows the user to create levels with a level editor.  
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "level_edit_state.h"

#include "marker_lost_state.h"
#include "marker_select_state.h"
#include "play_state.h"

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Init
Purpose : Initialises the level editor and creates a test level.
====================================================================================*/
void LevelEditState::Init()
{
	level_editor_.Init(platform_, level_blocks_);

	if (marker_index_ == 0)
	{
		CreateTestLevel();
	}
}

/*====================================================================================
Name    : CleanUp
Purpose : Performs any necessary memory cleanup.  
====================================================================================*/
void LevelEditState::CleanUp()
{
}

/*====================================================================================
Name    : Update
Purpose : Handles the main loop for this state.  
====================================================================================*/
void LevelEditState::Update(float frame_time, const abfw::SonyController* controller, StateManager* state_manager)
{
	// if the ar marker isn't found through the camera
	if (sampleIsMarkerFound(marker_index_) == false)
	{
		// create a new marker lost state
		MarkerLostState* marker_lost_state;
		marker_lost_state = new MarkerLostState(platform_);

		// set the marker index on the state
		marker_lost_state->set_marker_index(marker_index_);
		marker_lost_state->set_player_marker_index(marker_index_);

		// pause the level edit state and change to the marker lost state
		state_manager->PushState(marker_lost_state);
	}
	// if the ar marker is found
	else
	{
		// update the level editor
		level_editor_.Update(frame_time, controller, level_blocks_);

		// transform the level editor's local transform by the marker transform
		level_editor_.MultiplyByMarkerTransform(marker_index_);

		// loop through all the level blocks
		for (std::vector<GameObject>::iterator block = level_blocks_.begin(); block < level_blocks_.end(); block++)
		{
			// transform the current block's local transform by the marker transform
			block->MultiplyByMarkerTransform(marker_index_);
		}

		// go back to the marker select state when the select button is pressed
		if (controller->buttons_pressed() & ABFW_SONY_CTRL_SELECT)
		{
			state_manager->ChangeState(new MarkerSelectState(platform_));
		}

		// if the start button is pressed
		if (controller->buttons_pressed() & ABFW_SONY_CTRL_START)
		{
			// create a new play state
			PlayState* play_state;
			play_state = new PlayState(platform_);

			// set the marker index on the state
			play_state->set_marker_index(marker_index_);

			// pass the level to the play state
			play_state->set_level_blocks(level_blocks_);

			// pause the level edit state and change to the play state
			state_manager->PushState(play_state);
		}
	}
}

/*====================================================================================
Name    : Render
Purpose : Renders the level blocks and the level editor.  
====================================================================================*/
void LevelEditState::Render(abfw::SpriteRenderer* sprite_renderer, abfw::Renderer3D* renderer_3d)
{
	// loop through the vector backwards so that the geometry is rendered before the start point and goal
	for (std::vector<GameObject>::reverse_iterator block = level_blocks_.rbegin(); block < level_blocks_.rend(); block++)
	{
		block->Render(renderer_3d);
	}

	level_editor_.Render(renderer_3d);
}

/*====================================================================================
Name    : CreateTestLevel
Purpose : Creates a test level for the user to play on.  
====================================================================================*/
void LevelEditState::CreateTestLevel()
{
	// create 3 long steps for the player to jump up
	for (int step = 0; step < 3; step++)
	{
		for (int x_position = 2; x_position < 7; x_position++)
		{
			level_editor_.CreateBlock(x_position * kBlockSize, (step + 2) * kBlockSize, step * kBlockSize, level_blocks_, kLevelGeometry);
		}
	}

	// place the start point on the left of the bottom step
	// and place the goal on the right of the top step
	level_blocks_[0].SetPosition(2 * kBlockSize, 2 * kBlockSize, kBlockSize);
	level_blocks_[1].SetPosition(6 * kBlockSize, 4 * kBlockSize, 3 * kBlockSize);
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
