/*====================================================================================
Class Name  : LevelEditor
Authors	    : Ewan Duncan
Description : A movable cursor that can create and delete cubes that make up the
			  geometry of a playable level.
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "level_editor.h"

#include <input/sony_controller_input_manager.h>

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Init
Purpose : Performs necessary set up for the level editor.
====================================================================================*/
void LevelEditor::Init(abfw::Platform* platform, std::vector<GameObject>& level_blocks)
{
	// store the platform for creating meshes
	platform_ = platform;

	// create a transparent blue cube that acts as the cursor
	set_mesh(CreateCubeMesh("glass.png", *platform));
	SetPosition(0.f, 0.f, 0.f);
	Scale(kBlockSize + 0.001f, kBlockSize + 0.001f, kBlockSize + 0.001f);
	MoveDeltaX(kBlockSize * 2);
	MoveDeltaY(kBlockSize * 2);

	// create meshes to be used for the level blocks
	start_block_ = CreateCubeMesh("start.png", *platform_);
	goal_block_ = CreateCubeMesh("goal.png", *platform_);
	geometry_block_ = CreateCubeMesh("block.png", *platform_);

	// create the start point and goal for the level and place them off screen
	CreateBlock(1000, 1000, 1000, level_blocks, kStart);
	CreateBlock(1000, 1000, 1000, level_blocks, kGoal);
}

/*====================================================================================
Name    : Update
Purpose : Calls functions to move the cursor and edit the level.  
====================================================================================*/
void LevelEditor::Update(float frame_time, const abfw::SonyController* controller, std::vector<GameObject>& level_blocks)
{
	MoveCursor(frame_time, controller);

	EditLevel(controller, level_blocks);
}

/*====================================================================================
Name    : MoveCursor
Purpose : Moves the cursor in 6 directions using the d-pad and shoulder buttons.  
====================================================================================*/
void LevelEditor::MoveCursor(float frame_time, const abfw::SonyController* controller)
{
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_LEFT)
	{
		MoveDeltaX(-kBlockSize);
	}
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_RIGHT)
	{
		MoveDeltaX(kBlockSize);
	}
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_UP)
	{
		MoveDeltaY(kBlockSize);
	}
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_DOWN)
	{
		MoveDeltaY(-kBlockSize);
	}
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_L2)
	{
		MoveDeltaZ(-kBlockSize);
	}
	if (controller->buttons_pressed() & ABFW_SONY_CTRL_R2)
	{
		MoveDeltaZ(kBlockSize);
	}
}

/*====================================================================================
Name    : EditLevel
Purpose : Lets the user place and delete blocks at the position of the cursor.  Also 
		  lets the user move the level start point and goal.  
====================================================================================*/
void LevelEditor::EditLevel(const abfw::SonyController* controller, std::vector<GameObject>& level_blocks)
{
	// flag for checking if a block is already placed at the position of the cursor
	bool blockPlacedHere = false;

	// flag for checking if a block is level geometry or not
	bool blockIsGeometry = false;

	// use to store the index of a block in the vector which is at the position of the cursor
	int placed_block_index;

	// loop through all the blocks in the vector of level blocks
	for (int block_index = 0; block_index < level_blocks.size(); block_index++)
	{
		// if the cursor is colliding with a block
		if (AABBCollision(level_blocks[block_index]) == true)
		{
			// note that a block is placed here and store the vector index
			blockPlacedHere = true;
			placed_block_index = block_index;

			// check if the block is level geometry
			if (level_blocks[block_index].get_object_type() == kLevelGeometry)
			{
				blockIsGeometry = true;
			}
		}
	}

	// if the x button is pressed
	if (controller->buttons_released() & ABFW_SONY_CTRL_CROSS)
	{
		// if there isn't a block already placed at this point, place one
		if (blockPlacedHere == false)
		{
			CreateBlock(GetPosition().x, GetPosition().y, GetPosition().z, level_blocks, kLevelGeometry);
		}
		// if there is a block already placed here
		else
		{
			// if the block is level geometry, then remove it
			if (blockIsGeometry == true)
			{
				level_blocks.erase(level_blocks.begin() + placed_block_index);
			}
		}
	}

	// if the square button is pressed
	if (controller->buttons_released() & ABFW_SONY_CTRL_SQUARE)
	{
		// if there isn't a block here, then move the level start point to this position
		if (blockPlacedHere == false)
		{
			level_blocks[0].SetPosition(GetPosition());
		}
	}

	// if the circle button is pressed
	if (controller->buttons_released() & ABFW_SONY_CTRL_CIRCLE)
	{
		// if there isn't a block here, then move the level goal to this position
		if (blockPlacedHere == false)
		{
			level_blocks[1].SetPosition(GetPosition());
		}
	}
}

/*====================================================================================
Name    : Createblock
Purpose : Creates a level block of a specified type at a chosen position and adds 
		  it to the vector of level blocks.  
====================================================================================*/
void LevelEditor::CreateBlock(float x_position, float y_position, float z_position, std::vector<GameObject>& level_blocks, ObjectType object_type)
{
	// create a game object to represent the block
	GameObject block;
	
	// check the block type and create an appropriate cube
	if (object_type == kLevelGeometry)
	{
		block.set_object_type(kLevelGeometry);
		block.set_mesh(geometry_block_);
		block.Scale(kBlockSize, kBlockSize, kBlockSize);
	}
	else if (object_type == kStart)
	{
		block.set_object_type(kStart);
		block.set_mesh(start_block_);
		block.Scale(kBlockSize - 0.001f, kBlockSize - 0.001f, kBlockSize - 0.001f);
	}
	else if (object_type == kGoal)
	{
		block.set_object_type(kGoal);
		block.set_mesh(goal_block_);
		block.Scale(kBlockSize - 0.001f, kBlockSize - 0.001f, kBlockSize - 0.001f);
	}

	// move the block to the specified position and add it to the vector
	block.SetPosition(x_position, y_position, z_position);
	level_blocks.push_back(block);
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
