/*====================================================================================
Class Name  : PlayState
Authors	    : Ewan Duncan
Description : Allows the user to play levels created in the editor.
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "play_state.h"

#include "marker_lost_state.h"
#include "level_edit_state.h"
#include "level_complete_state.h"

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Init
Purpose : 
====================================================================================*/
void PlayState::Init()
{
	player_.Init(platform_, level_blocks_, marker_index_);

	marker_was_lost_ = false;
}

/*====================================================================================
Name    : CleanUp
Purpose : Performs any necessary memory cleanup.
====================================================================================*/
void PlayState::CleanUp()
{
}

/*====================================================================================
Name    : Update
Purpose : Handles the main loop for this state.
====================================================================================*/
void PlayState::Update(float frame_time, const abfw::SonyController* controller, StateManager* state_manager)
{
	// if the ar markers aren't found through the camera
	if (sampleIsMarkerFound(marker_index_) == false || sampleIsMarkerFound(player_.get_player_marker_index()) == false)
	{
		marker_was_lost_ = true;

		// create a new marker lost state
		MarkerLostState* marker_lost_state;
		marker_lost_state = new MarkerLostState(platform_);

		// set the marker index on the state
		marker_lost_state->set_marker_index(marker_index_);
		marker_lost_state->set_player_marker_index(player_.get_player_marker_index());

		// pause the level edit state and change to the marker lost state
		state_manager->PushState(marker_lost_state);
	}
	// if the ar markers are found
	else
	{
		// if the marker was lost in the previous frame
		if (marker_was_lost_ == true)
		{
			// overwrite the previous marker position with the current marker position so that
			// the player doesn't teleport whenever the marker is found after being lost
			player_.set_previous_marker_position(player_.FindMarkerPosition());
			marker_was_lost_ = false;
		}

		// update the player
		player_.Update(frame_time, controller, level_blocks_);

		// loop through all the level blocks
		for (std::vector<GameObject>::iterator block = level_blocks_.begin(); block < level_blocks_.end(); block++)
		{
			// transform the current block's local transform by the marker transform
			block->MultiplyByMarkerTransform(marker_index_);
		}

		// if the start button is pressed, go back to the level edit state
		if (controller->buttons_pressed() & ABFW_SONY_CTRL_START)
		{
			state_manager->PopState();
		}

		// if the player reached the goal, change to the level complete state
		if (player_.get_goal_reached() == true)
		{
			state_manager->ChangeState(new LevelCompleteState(platform_));
		}
	}
}

/*====================================================================================
Name    : Render
Purpose : Renders the level blocks and the level editor.
====================================================================================*/
void PlayState::Render(abfw::SpriteRenderer* sprite_renderer, abfw::Renderer3D* renderer_3d)
{
	player_.Render(renderer_3d);

	// loop through the vector backwards so that the geometry is rendered before the start point and goal
	for (std::vector<GameObject>::reverse_iterator block = level_blocks_.rbegin(); block < level_blocks_.rend(); block++)
	{
		block->Render(renderer_3d);
	}
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
