/*====================================================================================
Class Name  : Player
Authors	    : Ewan Duncan
Description : Object that represents the player in the game.  Can be moved with an 
			  ar marker and jump with button input.  
====================================================================================*/

/*------------------------------------------------------------------------------------
INCLUDES
------------------------------------------------------------------------------------*/
#include "player.h"

#include <input/sony_controller_input_manager.h>
#include <sony_sample_framework.h>

/*------------------------------------------------------------------------------------
MEMBER FUNCTIONS
------------------------------------------------------------------------------------*/

/*====================================================================================
Name    : Init
Purpose : Performs necessary set up for the player.
====================================================================================*/
void Player::Init(abfw::Platform* platform, std::vector<GameObject>& level_blocks, const int level_marker_index)
{
	// set marker indices for the level and player
	level_marker_index_ = level_marker_index;
	player_marker_index_ = 5;

	// player dimensions
	const float half_width = 5.f;
	const float height = 10.f;
	const float block_size = 0.025f;
	const float feet_height = 0.01f;

	// create the bounding box for the player
	abfw::Aabb aabb;
	aabb.set_min_vtx(abfw::Vector3(-half_width, 0.f, -half_width));
	aabb.set_max_vtx(abfw::Vector3(half_width, height, half_width));
	
	// load in the model and use its mesh
	set_mesh(CreateModelMesh("Walrus.obj", *platform, aabb));
	
	// rotate and scale the player to a desirable orientation and size
	SetXRotation(90.f);
	Scale(block_size * 0.1f, block_size * 0.1f, block_size * 0.1f);

	// move the player to the level start point
	SetPosition(level_blocks[0].GetPosition());

	// offset the player so it is standing on the level instead of hovering above
	// (necessary due to model not having its origin centered)
	MoveDeltaZ(-block_size * 0.5f);

	// set the previous position of the marker to match the marker position for later use
	previous_marker_position_ = FindMarkerPosition();

	// the feet game object is used for detecting when the player is standing on a platform
	// set it's mesh to be a box that covers the bottom of the model's feet
	feet_.set_mesh(CreateCubeMesh("glass.png", *platform));
	feet_.Scale(block_size, block_size, feet_height);
	feet_.SetPosition(GetPosition());

	// initialise other properties
	goal_reached_ = false;
	on_platform_ = false;
	vertical_velocity_ = 0.f;
}

/*====================================================================================
Name    : Update
Purpose : Calls functions to move the player, jump and check for collisions.
====================================================================================*/
void Player::Update(float frame_time, const abfw::SonyController* controller, std::vector<GameObject>& level_blocks)
{
	// store the current position of the player in case it collides with an object and needs to be moved back
	abfw::Vector3 temp_position = GetPosition();

	// use an ar marker to move the player along the x and y axis
	MoveWithMarker(frame_time);
	
	// check if the player is standing on a platform or not
	PlatformCollisionDetection(level_blocks);

	// update the player's vertical velocity and allow them to jump
	Jump(frame_time, controller);

	// checks for collisions with the level blocks
	LevelCollisionDetection(temp_position, level_blocks);

	// transform local transforms by marker transforms
	MultiplyByMarkerTransform(level_marker_index_);
	feet_.MultiplyByMarkerTransform(level_marker_index_);
}

/*====================================================================================
Name    : FindMarkerPosition
Purpose : Calculate the local marker transform of the player marker in relation to 
		  the level marker and return its position.
====================================================================================*/
abfw::Vector3 Player::FindMarkerPosition()
{
	return GetLocalMarkerTransform(level_marker_index_, player_marker_index_).GetTranslation();
}

/*====================================================================================
Name    : MoveWithMarker
Purpose : Moves the player along the x and y axis with the player ar marker.
====================================================================================*/
void Player::MoveWithMarker(float frame_time)
{
	// find the current position of the marker
	abfw::Vector3 current_marker_position = FindMarkerPosition();

	// calculate the x and y distances the player needs to move by finding the distance between the previous 
	// marker position and the current marker position
	// multiply the results by the frame time and scale them up by 100 so they're not too small
	float x_distance = (current_marker_position.x - previous_marker_position_.x) * 100.f * frame_time;
	float y_distance = (current_marker_position.y - previous_marker_position_.y) * 100.f * frame_time;

	// move the player by the calculated distances
	MoveDeltaX(x_distance);
	MoveDeltaY(y_distance);

	// set the previous marker position to the current marker position for use in the next frame
	previous_marker_position_ = current_marker_position;
}

/*====================================================================================
Name    : Jump
Purpose : Updates the player's vertical velocity and allows them to jump.
====================================================================================*/
void Player::Jump(float frame_time, const abfw::SonyController* controller)
{
	// when the player isn't standing on a platform, add a downwards acceleration onto
	// the vertical velocity to make the player fall
	if (on_platform_ == false)
	{
		vertical_velocity_ = vertical_velocity_ + kVerticalAcceleration;
	}

	// cap the vertical velocity so the player reaches a terminal velocity when falling
	if (vertical_velocity_ > kMaxVerticalVelocity)
	{
		vertical_velocity_ = kMaxVerticalVelocity;
	}

	// if the player is falling
	if (vertical_velocity_ >= 0.f)
	{
		// set their vertical velocity to zero when they land on a platform
		if (on_platform_ == true)
		{
			vertical_velocity_ = 0.f;
		}
	}

	// if the player is standing on a platform
	if (on_platform_ == true)
	{
		// make the player jump when the user press up on the d-pad
		if (controller->buttons_pressed() & ABFW_SONY_CTRL_UP)
		{
			vertical_velocity_ = vertical_velocity_ - kInitialVerticalVelocity;
			on_platform_ = false;
		}
	}

	// move the player by the vertical velocity
	MoveDeltaZ(-vertical_velocity_ * frame_time);

	// update the position of the feet object so it mains at the feet of the player
	feet_.SetPosition(GetPosition());
}

/*====================================================================================
Name    : PlatformCollisionDetection
Purpose : Checks if a player is standing on a platform or not.  
====================================================================================*/
void Player::PlatformCollisionDetection(std::vector<GameObject>& level_blocks)
{
	// flag for if a collision occurred
	bool collision_occurred = false;

	// loop through all the level geometry in the vector of level blocks
	for (int block_index = 2; block_index < level_blocks.size(); block_index++)
	{
		// if the feet are colliding with a block
		if (feet_.AABBCollision(level_blocks[block_index]) == true)
		{
			// flag that a collision occurred
			collision_occurred = true;
		}
	}

	// set the on_platform flag
	on_platform_ = collision_occurred;
}

/*====================================================================================
Name    : LevelCollisionDetection
Purpose : Checks if the player is colliding with a level block and if the player 
		  has fallen off the level.  
====================================================================================*/
void Player::LevelCollisionDetection(abfw::Vector3 temp_position, std::vector<GameObject>& level_blocks)
{
	// loop through all the level geometry in the vector of level blocks
	for (int block_index = 2; block_index < level_blocks.size(); block_index++)
	{
		// if the player is moving through level geometry
		if (AABBCollision(level_blocks[block_index]) == true)
		{
			// move the player back to stop it moving through the block
			SetPosition(temp_position);
		}
	}

	// if the player falls below the level, move them back to the start point
	if (GetPosition().z < -0.5f)
	{
		SetPosition(level_blocks[0].GetPosition());
	}

	// flag when the player reaches the goal
	if (AABBCollision(level_blocks[1]) == true)
	{
		goal_reached_ = true;
	}
}

/*------------------------------------------------------------------------------------
END OF FILE
------------------------------------------------------------------------------------*/
